package tas_sample;

import org.eclipse.paho.client.mqttv3.*;
import org.eclipse.paho.client.mqttv3.persist.MemoryPersistence;
import org.json.JSONObject;

import java.nio.charset.StandardCharsets;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.TimeUnit;

/*
範例程式說明：
    撥出電話詢問用戶到餐廳用餐的滿意度。

程式使用的外部函式庫，記載於 build.gradle.kts 檔
目前使用下列函式庫：
* [JSON In Java](https://mvnrepository.com/artifact/org.json/json)
* [Eclipse Paho Client Mqttv3](https://mvnrepository.com/artifact/org.eclipse.paho/org.eclipse.paho.client.mqttv3)

HTTP POST callout API 輸入 JSON 內容:
{
  "serviceNumber": "填入租用的服務代表號",
  "phones": [
    "填入被通知的電話號碼"
  ],
  "ivrData": {
    "node": "MAIN",
    "nextNode": "CUSTOM"
  }
}

詳細說明文件：
    https://tasapi.cht.com.tw/tas/download/tas_api_20191109.pdf
*/

public class CalloutMQTTSample {

    //tasapi.cht.com.tw TAS API 入口
    private static final String brokerUrl = "ssl://tasapi.cht.com.tw:2883";
    //寫入您的 X-API-KEY 及 SNKey
    private static final String MakerAPIKey = "YOUR_X_API_KEY";
    //SNKey 可由 HTTP GET TAS API https://tasapi.cht.com.tw/apis/CHTIoT/phone-conn/v1/reg 查詢
    private static final String SNKey = "YOUR_SNKEY";

    //MQTT 訊息 topic 列表
    private static final String calloutResultTopic = "phone-conn/calloutResult/" + SNKey;
    private static final String callEventTopic = "phone-conn/callEvent/" + SNKey;
    private static final String callActionTopic = "phone-conn/callAction/" + SNKey;
    private static final String callActionDebugTopic = "phone-conn/callActionDebug/" + SNKey;

    //發送訊息的 thread
    private final ScheduledExecutorService workerService = Executors.newScheduledThreadPool(1);
    private final MqttClient mqttClient;
    private final MqttConnectOptions connOpts;

    //建構子
    public CalloutMQTTSample() throws MqttException {
        //初始化 MQTT 連線參數
        connOpts = new MqttConnectOptions();
        connOpts.setUserName(MakerAPIKey);
        connOpts.setPassword(MakerAPIKey.toCharArray());
        //不接收前次 session 的訊息
        connOpts.setCleanSession(true);
//        connOpts.setKeepAliveInterval(60); //default

        String mqttClientId = MqttClient.generateClientId();
        mqttClient = new MqttClient(brokerUrl, mqttClientId, new MemoryPersistence());
    }

    //程式進入點
    public static void main(String[] args) throws Exception {
        CalloutMQTTSample calloutMQTTSample = new CalloutMQTTSample();
        calloutMQTTSample.connectMQTTServer();
    }

    //發送訊息到 TAS 伺服器
    public void publishMsg(String msg) {
        workerService.submit(() -> {
            try {
                mqttClient.publish(callActionTopic, msg.getBytes(StandardCharsets.UTF_8), 0, false);
            } catch (MqttException e) {
                e.printStackTrace();
            }
        });
    }

    //處理從 MAIN 節點回應的訊息
    private void handleMain(JSONObject req) {
        //建立要送到 CallAction topic 的 JSON 訊息
        JSONObject resp = new JSONObject();

        //送出訊息的 id 值必須與收到訊息的 id 值完全相同
        resp.put("id", req.get("id"));
        //自訂節點名稱，後續方便 if-else 或 switch-case 判斷節點名稱以進行後續處理
        resp.put("node", "askNode");
        //撥放語音
        resp.put("text", "您好，請問近期到阿明餐館的消費體驗，滿分5分，請問您給幾分評價");
        //使用 女音
        resp.put("promptMode", "F");
        //開啟接收用戶輸入按鍵
        resp.put("collectDTMF", true);
        //最多輸入 1 碼數字按鍵
        resp.put("maxDTMF", "1");
        //設定 3 秒沒有輸入判定為 timeout
        resp.put("collectDTMFTimeout", "3");
        //CUSTOM 代表後續也是由 MQTT 訊息處理
        resp.put("nextNode", "CUSTOM");

        //送出控制訊息
        publishMsg(resp.toString());
    }

    //askNode 節點的回應處理程式
    private void askNode(JSONObject req) {
        String inputDigitStr = req.getString("node");
        String digit = "";
        //取得回應訊息中用戶按下的數字按鍵
        if (inputDigitStr.contains(",")) {
            String[] splitArray = inputDigitStr.split(",");
            if (splitArray.length == 2) {
                digit = splitArray[1].trim();
            }
        }
        //印出收到的 DTMF 訊息
        System.out.println("input = " + digit);

        //設定只允許輸入數字 1~5 才視為正確輸入
        switch (digit) {
            case "1":
            case "2":
            case "3":
            case "4":
            case "5":
                //處理輸入正常數字
                handleNormalInput(req);
                break;
            case "timeout":
                //用戶沒有任何輸入的情況
                handleTimeout(req);
                break;
            default:
                //輸入錯誤數字的情況
                handleError(req);
        }
    }

    private void handleNormalInput(JSONObject req) {
        //送回 MQTT callAction topic 的 id 值，要與本次從 callEvent
        //收到訊息的 id 值完全相同
        JSONObject resp = new JSONObject();
        resp.put("id", req.get("id"));
        resp.put("node", "bye");
        resp.put("text", "感謝您的評分，再見");
        resp.put("promptMode", "F");
        resp.put("nextNode", "END");

        publishMsg(resp.toString());
    }

    private void handleTimeout(JSONObject req) {
        //送回 MQTT callAction topic 的 id 值，要與本次從 callEvent
        //收到訊息的 id 值完全相同
        JSONObject resp = new JSONObject();
        resp.put("id", req.get("id"));
        resp.put("node", "askNode");
        resp.put("text", "沒收到您的輸入，請問近期到阿明餐館的消費體驗，滿分5分，請問您給幾分評價");
        resp.put("promptMode", "F");
        resp.put("collectDTMF", true);
        resp.put("maxDTMF", "1");
        resp.put("collectDTMFTimeout", "3");
        resp.put("nextNode", "CUSTOM");

        publishMsg(resp.toString());
    }

    private void handleError(JSONObject req) {
        //送回 MQTT callAction topic 的 id 值，要與本次從 callEvent
        //收到訊息的 id 值完全相同
        JSONObject resp = new JSONObject();
        resp.put("id", req.get("id"));
        resp.put("node", "askNode");
        resp.put("text", "輸入錯誤，請輸入數字 1 到 5，請問近期到阿明餐館的消費體驗，您給予幾分評價");
        resp.put("promptMode", "F");
        resp.put("collectDTMF", true);
        resp.put("maxDTMF", "1");
        resp.put("collectDTMFTimeout", "3");
        resp.put("nextNode", "CUSTOM");

        publishMsg(resp.toString());
    }

    //連線 MQTT 伺服器及訂閱要接收訊息的 topics
    public void connectMQTTServer() {
        try {
            System.out.println("broker = " + brokerUrl);

            if (!mqttClient.isConnected()) {
                System.out.println("connecting MQTT broker...");
                mqttClient.connect(connOpts);
                System.out.println("connected");
            }

            //設定 MQTT client 程式收到訊息時的處理方式
            mqttClient.setCallback(new MqttCallback() {
                @Override
                public void connectionLost(Throwable cause) {
                    //MQTT 連線中斷時印出錯誤訊息
                    System.out.println("connectionLost() is called, reason: " + cause);
                    //連線中斷時，在 5 秒後重新連線 MQTT 伺服器
//                    tryReconnectMQTT();
                }

                @Override
                public void messageArrived(String topic, MqttMessage message) {
                    //印出從 topic 收到的 MQTT 訊息
                    String msg = new String(message.getPayload(), StandardCharsets.UTF_8);
                    System.out.println("global msg: " + msg);
                }

                @Override
                public void deliveryComplete(IMqttDeliveryToken token) {
                    //印出已送出到 TAS 服務的訊息
//                System.out.println("deliveryComplete() is called");
                    try {
                        String msg = new String(token.getMessage().getPayload(), StandardCharsets.UTF_8);
                        JSONObject jsonObject = new JSONObject(msg);
                        System.out.println("sent msg: " + jsonObject.toString(2) + "\n");
                    } catch (MqttException e) {
                        e.printStackTrace();
                    }
                }
            });

            //針對收到從 calloutResult topic 訊息的處理程式
            mqttClient.subscribeWithResponse(calloutResultTopic, 0, (topic, message) -> {
                String msg = new String(message.getPayload(), StandardCharsets.UTF_8);
                System.out.println("[calloutResult] received msg: " + msg + "\n");
            });


            //針對收到從 callEvent topic 訊息的處理程式
            mqttClient.subscribeWithResponse(callEventTopic, 0, (topic, message) -> {
                String msg = new String(message.getPayload(), StandardCharsets.UTF_8);
                System.out.println("[callEvent] received msg: " + msg + "\n");

                //handle message
                JSONObject req = new JSONObject(msg);
                String node = req.optString("node");
                String nodeName = node.split(",")[0].trim();

                //依 node 名稱執行對應的處理函式
                switch (nodeName) {
                    case "MAIN":
                        handleMain(req);
                        break;
                    case "askNode":
                        askNode(req);
                        break;
                }
            });

            //針對收到從 callActionDebug topic 訊息的處理程式，接收伺服器回應的 debug 用訊息
            mqttClient.subscribeWithResponse(callActionDebugTopic, 0, (topic, message) -> {
                String msg = new String(message.getPayload(), StandardCharsets.UTF_8);
                System.out.println("[callActionDebug] debug msg: " + msg);
            });

        } catch (MqttException ex) {
            System.out.println("code: " + ex);
            System.out.println("MqttException: " + ex);

            //發生問題時，延遲 5 秒後重新連線 MQTT server,
//            tryReconnectMQTT();
        } catch (Exception ex) {
            System.out.println("Exception: " + ex);
        }
    }

    /**
     * 延遲幾秒後，重新連線 MQTT 伺服器
     */
    private void tryReconnectMQTT() {
        workerService.schedule(this::connectMQTTServer, 5, TimeUnit.SECONDS);
    }

    //關閉 MQTT 連線
    private void shutdown(MqttClient client) {
        try {
            if (client != null) {
                if (client.isConnected()) {
                    client.disconnect();
                }
                client.close();
            }

        } catch (MqttException e1) {
            e1.printStackTrace();
        }
    }
}
